/**************************************************************
 * The Sound class
 * By: Shaun M. McFall
 * Under: Dr. Krone, Dr. Bressoud
 * Summer, 2007
 **************************************************************/

import javax.sound.sampled.*;
import java.io.*;
import java.lang.IllegalArgumentException;

public class Sound
{
  // The filename of the sound
  private String fileName = null;
  // Default sample rate, half the rate of a CD
  private static int SAMPLE_RATE = 22050;
  // Default sample size
  private static int SAMPLE_SIZE = 16;
  // An array of bytes to represent the sound
  private byte[] buffer;
  // Contains the "file type, the file's length in bytes, 
  //  the length in sample frames of the audio data contained 
  //  in the file, and the format of the audio data"
  private AudioFileFormat audioFileFormat = null;
  // The number of frames and bytes
  private int frames, bytes;
  // This is the Play object that allows for palying
  private Play play = null;
  
  // Constructors
  /**
   * Constructs a <code>Sound</code> of the specified length.
   * This sound will simply consist of an empty byte array, and an
   * <code>AudioFileFormat</code> with the following values:
   * <ul> 
   * <li><code>AudioFileFormat.Type.WAVE</code>
   * <li>22.05K sampling rate
   * <li>16 bit sample
   * <li>1 channel
   * <li>signed PCM encoding
   * <li>small-endian byte order
   * </ul>
   * Note that no new sound file is created, we only represent the sound
   * with a buffer and the AudioFileFormat.
   *
   * @param sec the length of the sound in seconds
   */
  public Sound(double sec) {
    int numChan = 1; // How many channels of sound there will be
    int bytesPerSample = SAMPLE_SIZE / 8; // bps
    
    // Make a new sound with 22k rate, 16 bits, 1 channel,
    //  signed, little-endian order.
    AudioFormat format = new AudioFormat(SAMPLE_RATE, SAMPLE_SIZE, 
                                         numChan, true, false);
    
    // The length of the sound in frames and in bytes
    frames = (int)(sec * SAMPLE_RATE * numChan);
    bytes = frames * bytesPerSample;
    
    // Create the WAV
    audioFileFormat = new AudioFileFormat(AudioFileFormat.Type.WAVE, 
                                          format, frames);
    
    // Create the buffer
    buffer = new byte[bytes];
  }
  
  /**
   * Constructs a <code>Sound</code> of the specified length.
   * This sound will simply consist of an empty byte array, and an
   * <code>AudioFileFormat</code> with the following values:
   * <ul> 
   * <li><code>AudioFileFormat.Type.WAVE</code>
   * <li>22.05K sampling rate
   * <li>16 bit sample
   * <li>1 channel
   * <li>signed PCM encoding
   * <li>small-endian byte order
   * </ul>
   * Note that no new sound file is created, we only represent the sound
   * with a buffer and the AudioFileFormat.
   *
   * @param numFrames the length of the sound in frames
   */
  public Sound(int numFrames) {
    int numChan = 1; // How many channels of sound there will be
    int bytesPerSample = SAMPLE_SIZE / 8; // bps
    
    // Make a new sound with 22k rate, 16 bits, 1 channel,
    //  signed, little-endian order.
    AudioFormat format = new AudioFormat(SAMPLE_RATE, SAMPLE_SIZE, 
                                         numChan, true, false);
    
    // The length of the sound in frames and in bytes
    frames = numFrames * numChan;
    bytes = frames * bytesPerSample;
    
    // Create the WAV
    audioFileFormat = new AudioFileFormat(AudioFileFormat.Type.WAVE, 
                                          format, frames);
    // Create the buffer
    buffer = new byte[bytes];
  }
  
  /**
   * Constructs a <code>Sound</code> from the given file name.
   *
   * @param path the path of the file to read in
   */
  public Sound(String path) {
    // Create a File object from the given path
    File file = new File(path);
    // Create an AudioInputStream, which allows us to read it in
    AudioInputStream inputStream;
    try {
      inputStream = AudioSystem.getAudioInputStream(file);
    } catch(Exception e) {
      System.err.println("Unable to read from file " + 
                         path + ".  The file type is unsupported.  " + 
                         "Are you sure you're using a WAV, AU, or " +
                         "AIFF file (some .wav files are encoded " +
                         "using mp3)?");
      return;
    }//catch
    // Check to make sure we're not too big (from SimpleSound)
    if((inputStream.getFrameLength() * 
        inputStream.getFormat().getFrameSize()) > Integer.MAX_VALUE) {
      System.err.println("The sound in file: " + path + " is too long."+
                 "  Try using a shorter sound.");
    }
    // Get the size of the stream
    int bufferSize = (int)inputStream.getFrameLength() * inputStream.getFormat().getFrameSize();
    buffer = new byte[bufferSize];
    bytes = bufferSize;
    frames = (int)inputStream.getFrameLength();
    
    int off = 0;
    int numRead = 0;
    // Read in the data (from SimpleSound, but the most
    //  efficient way to do it anyways).
    while (true) {
      try {
        numRead = inputStream.read(buffer, off, bufferSize);
        if(numRead == -1)//no more data
          break;
        else
          off += numRead;
      } catch(Exception e) {
        System.err.println("Error reading in data!");
        return;
      }
    }
    
    // Get the file type, once again borrowed from SimpleSound,
    //  but also really the only way to do it...
    // Set the format of the file, assuming that the extension is correct
    if(path.toLowerCase().endsWith(".wav")) {
      audioFileFormat = 
        new AudioFileFormat(AudioFileFormat.Type.WAVE,
                            inputStream.getFormat(),
                            (int)inputStream.getFrameLength());
    } else if(path.toLowerCase().endsWith(".au")) {
      audioFileFormat = 
        new AudioFileFormat(AudioFileFormat.Type.AU,
                            inputStream.getFormat(),
                            (int)inputStream.getFrameLength());
    } else if (path.toLowerCase().endsWith(".aif")||
             path.toLowerCase().endsWith(".aiff")) {
      audioFileFormat = 
        new AudioFileFormat(AudioFileFormat.Type.AIFF,
                            inputStream.getFormat(),
                            (int)inputStream.getFrameLength());
    } else {
      System.err.println("Unsupported file type.  Please try again with a "+
                 "file that ends in .wav, .au, .aif, or .aiff");
    }
    
    // Update the file name
    fileName = path;
    // This takes the sound, and turns it to our default format
    Sound newSound = new Sound(frames);
    newSound.setSamples(getSamples());
    buffer = newSound.buffer;
    bytes = newSound.bytes;
    audioFileFormat = newSound.audioFileFormat;
  }
  
  /**
   * Creates an <code>AudioInputStream</code> for this sound from the 
   * <code>buffer</code> and the <code>audioFileFormat</code>.
   * @return an AudioInputStream representing this sound.
   * @see AudioInputStream
   */
  public AudioInputStream makeAIS() {
    AudioFileFormat.Type fileType = audioFileFormat.getType();
    ByteArrayInputStream bais = new ByteArrayInputStream(buffer);
    int frameSize = audioFileFormat.getFormat().getFrameSize();
    
    AudioInputStream audioInputStream = 
      new AudioInputStream(bais, audioFileFormat.getFormat(), 
                           buffer.length/frameSize);
    return audioInputStream;
  }
  
  // Play methods
  
  /**
   * Plays the sound object by adding it to the Play thread.
   * 
   * Note: This method will play the sound, but it will not wait
   * until the sound is done to continue program execution.
   */
  public void play() {
    // Clip clip = makeClip();
    // clip.start();
    play = new Play(this);
    play.start();
  }
  
  /**
   * Plays the sound object by adding it to the Play thread, and
   * waits until playback is done in order to continue program
   * execution.
   */
  public void blockPlay() {
    play = new Play(this);
    play.start();
    while(play.isAlive()) {
    
    }
  }
  
  /**
   * Plays the sound object by adding it to the Play thread, if
   * the method gets called again, it will not play it a second
   * time.
   * 
   * Note: This method is still unstable.
   */
  public void playOnce() {
    if (play != null) {
      if (play.getLeft() > 0)
        return;
    }
    play = new Play(this);
    play.start();
    //while(!play.isPlaying()) {}
  }
  
  // Methods to get sample
  /**
   * Method to get the sample at a given index.
   * 
   * Note: This will call getSample(), this is
   * because we will eventuall want that to throw
   * an exception (possibly).
   * 
   * @param index the index of the sample
   * @return the sample
   */
  private int getSampleAt(int index) {
    // Eventually we'll want this to throw an error
    int sample = getSample(index);
    return sample;
  }
  
  /**
   * Method to get all the samples of the sound
   * 
   * @return the samples
   */
  public int[] getSamples() {
    int[] samp = new int[frames];
    for (int i = 0; i < frames; i++)
      samp[i] = getSample(i);
    return samp;
  }
  
  /**
   * Method to get the sample at a given index.
   * 
   * Note: This is called by getSampleAt() because
   * we will want exception handling.
   * 
   * @param index the index of the sample
   * @return the sample
   */
  // If mono get the sample at the frame, if stereo get
  //  the leftmost
  public int getSample(int index) {
    // Check to make sure the fram is there
    if(index >= audioFileFormat.getFrameLength()) {
      throw new IllegalArgumentException("You are trying to access the sample at index: " 
                                           + (index) + ", but the last valid index is at " + 
                                         (audioFileFormat.getFrameLength() - 1));
    } else if(index < 0) {
      throw new IllegalArgumentException("You asked for the sample at index: " + (index) +
                         ".  This number is less than zero.  Please try" +
                         "again using an index in the range [0," + 
                         (audioFileFormat.getFrameLength() - 1) +"]");
    }
    
    // Now get the data we need
    AudioFormat format = audioFileFormat.getFormat(); // The format we're in
    int bits = format.getSampleSizeInBits(); // The sample size bits
    boolean isBigEndian = format.isBigEndian(); // How it it ordered?
    
    int frameSize = format.getFrameSize();
    byte[] theFrame = new byte[frameSize];
    for (int i = 0; i < frameSize; i++)
      theFrame[i] = buffer[index * frameSize + i];
    
    // From this point on we use SimpleSound stuff, but it's
    //  once again the only way to do it, if I typed it, it would
    //  look almost identical.
    if(format.getEncoding().equals(AudioFormat.Encoding.PCM_SIGNED))
    {
      //since we're always returning the left sample, 
      //we don't care if we're mono or stereo, left is
      //always first in the frame
      if(bits == 8)//8 bits == 1 byte
        return theFrame[0];
      else if(bits == 16)
        return TConversionTool.bytesToInt16(theFrame, 0, isBigEndian);
      else if(bits == 24)
        return TConversionTool.bytesToInt24(theFrame, 0, isBigEndian);
      else if(bits == 32)
        return TConversionTool.bytesToInt32(theFrame, 0, isBigEndian);
      else
      {
        System.err.println("Unsupported audio encoding.  The sample "+
                   "size is not recognized as a standard "+ 
                   "format.");
        return -1;
      }
    }
    else if(format.getEncoding().equals(AudioFormat.Encoding.PCM_UNSIGNED))
    {
      if(bits == 8)
        return TConversionTool.unsignedByteToInt(theFrame[0])-
        (int)Math.pow(2,7);
      else if(bits == 16)
        return TConversionTool.unsignedByteToInt16(theFrame, 0, isBigEndian)-
        (int)Math.pow(2, 15);
      else if(bits == 24)
        return TConversionTool.unsignedByteToInt24(theFrame, 0, isBigEndian)-
        (int)Math.pow(2, 23);
      else if(bits == 32)
        return TConversionTool.unsignedByteToInt32(theFrame, 0, isBigEndian)-
        (int)Math.pow(2, 31);
      else
      {
        System.err.println("Unsupported audio encoding.  The sample "+
                   "size is not recognized as a standard "+
                   "format.");
        return -1;
      }
    }
    else if(format.getEncoding().equals(AudioFormat.Encoding.ALAW))
    {
      return TConversionTool.alaw2linear(buffer[0]);
    }
    else if(format.getEncoding().equals(AudioFormat.Encoding.ULAW))
    {
      return TConversionTool.ulaw2linear(buffer[0]);
    }
    else
    {
      System.err.println("unsupported audio encoding: " + 
                 format.getEncoding() + ".  Currently only PCM, " +
                 "ALAW and ULAW are supported.  Please try again" +
                 "with a different file.");
      return -1;
    }
  }
  
  /**
   * Changes the value of each byte of the specified frame.
   * 
   * @param frameNum the index of the frame to change
   * @param theFrame the byte array that will be copied into this sound's
   *                 buffer in place of the specified frame.
   */
  private void setFrame(int frameNum, byte[] theFrame) {
    if(frameNum >= audioFileFormat.getFrameLength()) {
      System.err.println("That frame, number "+frameNum+", does not exist. "+
                 "The last valid frame number is " + 
                 (audioFileFormat.getFrameLength() - 1));
    }
    int frameSize = audioFileFormat.getFormat().getFrameSize();
    if(frameSize != theFrame.length)
      System.err.println("Frame size doesn't match, line 383.  This should" +
                 " never happen.  Please report the problem to a TA.");
    for(int i = 0; i < frameSize; i++) {
      buffer[frameNum*frameSize+i] = theFrame[i];
    }
  }
  
  /**
   * Method to set the sample value at the passed index to the passed value
   * @param index the index
   * @param value the new value
   */
  private void setSampleAt(int index, int value) {
    setSample(index,value);
  }
  
  public void setSamples(int[] values) {
    for (int i = 0; i < frames; i++)
      setSample(i, values[i]);
  }
  
  /**
   * Changes the value of the sample found at the specified frame.  If this
   * sound has more than one channel, then this defaults to setting only the
   * first (left) sample.
   *
   * @param frameNum the index of the frame where the sample should be changed
   * @param sample an int representation of the new sample to put in this
   *               sound's buffer at the specified frame
   */
  public void setSample(int frameNum, int sample) {
    // Get the data we need
    if (frameNum < 0 || frameNum > frames)
      throw new IllegalArgumentException("You are trying to set a sample at: " + frameNum +
                                           ". Please only set between [0, " + frames +
                                         "].");
    AudioFormat format = audioFileFormat.getFormat(); // The format we're in
    int bits = format.getSampleSizeInBits(); // The sample size bits
    boolean isBigEndian = format.isBigEndian(); // How it it ordered?
    
    int frameSize = format.getFrameSize();
    byte[] theFrame = new byte[frameSize];
    for (int i = 0; i < frameSize; i++)
      theFrame[i] = buffer[frameNum * frameSize + i];
    // From here on out, just like in getting the samples, I used
    //  the method from SimpleSound, only because
    if(format.getEncoding().equals(AudioFormat.Encoding.PCM_SIGNED))
    {
      if(bits == 8)//8 bits = 1 byte = first cell in array
      {
        theFrame[0] = (byte)sample;
        setFrame(frameNum, theFrame);
      }
      else if(bits == 16)//2 bytes, first 2 cells in array
      {
        TConversionTool.intToBytes16(sample, theFrame, 0, isBigEndian);
        setFrame(frameNum, theFrame);
      }
      else if(bits == 24)
      {
        TConversionTool.intToBytes24(sample, theFrame, 0, isBigEndian);
        setFrame(frameNum, theFrame);
      }
      else if(bits == 32)
      {
        TConversionTool.intToBytes32(sample, theFrame, 0, isBigEndian);
        setFrame(frameNum, theFrame);
      }
      else
      {
        System.err.println("Unsupported audio encoding.  The sample"+
                   "size is not recognized as a standard format");
      }
    }//if format == PCM_SIGNED
    else if(format.getEncoding().equals(AudioFormat.Encoding.PCM_UNSIGNED))
    {
      if(bits == 8)
      {
        theFrame[0] = TConversionTool.intToUnsignedByte(sample);
        setFrame(frameNum, theFrame);
      }
      else if(bits == 16)
      {
        TConversionTool.intToUnsignedBytes16(sample, theFrame, 0, isBigEndian);
        setFrame(frameNum, theFrame);
      }
      else if(bits == 24)
      {
        TConversionTool.intToUnsignedBytes24(sample, theFrame, 0, isBigEndian);
        setFrame(frameNum, theFrame);
      }
      else if(bits == 32)
      {
        TConversionTool.intToUnsignedBytes32(sample, theFrame, 0, isBigEndian);
        setFrame(frameNum, theFrame);
      }
      
      else
      {
        System.err.println("Unsupported audio encoding.  The sample"+
                   " size is not recognized as a standard "+
                   "format.");
      }
    }
    else if(format.getEncoding().equals(AudioFormat.Encoding.ALAW))
    {
      if((sample>Short.MAX_VALUE)||(sample<Short.MIN_VALUE))
        System.err.println("You are trying to set the sample value to: "+
                   sample + ", but the maximum value for a sample"+
                   " in this format is: "+Short.MAX_VALUE+
                   ", and the minimum value is: "+Short.MIN_VALUE+
                   ".  Please choose a value in that range.");
      theFrame[0] = TConversionTool.linear2alaw((short)sample);
      setFrame(frameNum, theFrame);
    }
    else if(format.getEncoding().equals(AudioFormat.Encoding.ULAW))
    {
      
      if((sample>Short.MAX_VALUE)||(sample<Short.MIN_VALUE))
        System.err.println("You are trying to set the sample value to: "+
                   sample + ", but the maximum value for a sample"+
                   " in this format is: "+Short.MAX_VALUE+
                   ", and the minimum value is: "+Short.MIN_VALUE+
                   ".  Please choose a value in that range.");
      theFrame[0] = TConversionTool.linear2ulaw((short)sample);
      setFrame(frameNum, theFrame);
    }
    else
    {
      System.err.println("unsupported audio encoding: " + 
                 format.getEncoding() + ".  Currently only PCM, " +
                 "ALAW and ULAW are supported.  Please try again" +
                 "with a different file.");
    }
  }//setSample(int, int)
  
  /**
   * Method to get the number of frames in a sound
   * 
   * @return the number of frames in the sound
   */
  public int getNumSamples() {
    return frames;
  }
  
  /**
   * Returns whether or not the sound is playing
   * 
   * @return true - the sound is playing <br> false - the sound is not playing
   */
  public boolean isPlaying() {
    if (play != null)
      return play.isPlaying();
    return false;
  }
  
  /**
   * Writes the sound out to a file. <BR>
   * If you do not determine a file type, it assumes wav, and you can
   * only save in .wav format.
   * 
   * @param s the file name
   * @return true - the sound was saved <br> false - the sound did not save
   */
  public boolean writeFile(String s)
  {
    //If the pathname is null, the method discontinues and false is returned.
    if (s == null)
      return false;
    // Get the extention and determine what to do with it
    int dot = s.lastIndexOf(".");
    if (dot != -1) {
      String ext = s.substring(dot);
      if (!ext.equalsIgnoreCase(".wav")) {
        System.err.println("You can not save as anything but a .wav.");
        return false;
      }
    } else
      s += ".wav";
    // Create an AudioInputStream to represent the sound
    AudioInputStream ais = makeAIS();
    AudioFileFormat.Type format = audioFileFormat.getType();
    
    try { // reset the ais
      ais.reset();
    } catch(Exception e) {
      System.err.println("Unable to reset the audio stream.");
      return false;
    }
    // Get the file, and make a new one
    File file = new File(s);
    try {
      file.createNewFile();
    } catch(IOException e) {
      System.err.println("Cannot create the file.");
      return false;
    }
    // Now write the actual file
    try {
      if (AudioSystem.write(ais, format, file) == -1) {
        System.err.println("Cannot write the file.");
        return false;
      }
      fileName = s;
    } catch(IOException e) {
      System.err.println("Cannot create the file.");
      return false;
    }
    // Close the ais
    try {
      ais.close();
    } catch(Exception e) {
      System.err.println("Could not close ais.");
      return false;
    }
    return true;
  }
}